import ssl
import time
import urllib
from contextlib import contextmanager

import docker


# Because we wait for a go signal to come up, we might not be able to
# make the request right away:
def http_get(url, retries=10, poll=0.5):
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    for _i in range(retries):
        try:
            r = urllib.request.urlopen(url, context=ctx)  # noqa
            return r.read().decode("UTF-8")
        except (urllib.error.URLError, ConnectionResetError) as e:
            print("sleeping...")
            time.sleep(poll)
            error = e
    raise error


# Create a container, as a context manager.
# This container will be stopped and removed when the context manager exits.
@contextmanager
def create_container(image, **kwargs):
    container = docker.from_env().containers.create(image, detach=True, **kwargs)
    try:
        yield container
    finally:
        container.stop()
        container.remove()



import time
from os.path import join

import constellation
import docker
import yaml
from constellation import acme, docker_util
from psycopg2 import OperationalError, connect

from montagu_deploy import database


def montagu_constellation(cfg):
    proxy = proxy_container(cfg)
    containers = [
        api_container(cfg),
        db_container(cfg),
        admin_container(cfg),
        contrib_container(cfg),
        proxy,
        proxy_metrics_container(cfg),
        mq_container(cfg),
        flower_container(cfg),
        task_queue_container(cfg),
    ]

    if cfg.use_acme:
        acme_container = acme.acme_buddy_container(
            cfg.acme_config, "acme-buddy", proxy.name_external(cfg.container_prefix), "montagu-tls", cfg.hostname
        )
        containers.append(acme_container)

    if cfg.fake_smtp_ref:
        fake_smtp = fake_smtp_container(cfg)
        containers.append(fake_smtp)

    return constellation.Constellation(
        "montagu", cfg.container_prefix, containers, cfg.network, cfg.volumes, data=cfg, vault_config=cfg.vault
    )


def admin_container(cfg):
    name = cfg.containers["admin"]
    return constellation.ConstellationContainer(name, cfg.admin_ref)


def contrib_container(cfg):
    name = cfg.containers["contrib"]
    mounts = [
        constellation.ConstellationVolumeMount("templates", "/usr/share/nginx/html/templates"),
        constellation.ConstellationVolumeMount("guidance", "/usr/share/nginx/html/guidance"),
    ]
    return constellation.ConstellationContainer(name, cfg.contrib_ref, mounts=mounts)


def mq_container(cfg):
    name = cfg.containers["mq"]
    mounts = [
        constellation.ConstellationVolumeMount("mq", "/data"),
    ]
    return constellation.ConstellationContainer(name, cfg.mq_ref, mounts=mounts, ports=[cfg.mq_port])


def flower_container(cfg):
    name = cfg.containers["flower"]
    mq = cfg.containers["mq"]
    env = {
        "CELERY_BROKER_URL": f"redis://{mq}//",
        "CELERY_RESULT_BACKEND": f"redis://{mq}/0",
        "FLOWER_PORT": cfg.flower_port,
    }
    return constellation.ConstellationContainer(name, cfg.flower_ref, ports=[cfg.flower_port], environment=env)


def task_queue_container(cfg):
    name = cfg.containers["task_queue"]
    mounts = [
        constellation.ConstellationVolumeMount("burden_estimates", "/home/worker/burden_estimate_files"),
    ]
    return constellation.ConstellationContainer(name, cfg.task_queue_ref, configure=task_queue_configure, mounts=mounts)


def task_queue_configure(container, cfg):
    print("[task-queue] Configuring task-queue container")
    task_queue_config = {"host": cfg.containers["mq"], "servers": cfg.task_queue_servers, "tasks": cfg.task_queue_tasks}
    task_queue_config["servers"]["montagu"]["url"] = f"http://{cfg.containers['api']}:8080"
    if cfg.fake_smtp_ref:
        task_queue_config["servers"]["smtp"]["host"] = cfg.containers["fake_smtp"]
        task_queue_config["servers"]["smtp"]["port"] = 1025
    reports_cfg_filename = join(cfg.path, "diagnostic-reports.yml")
    with open(reports_cfg_filename) as ymlfile:
        diag_reports = yaml.safe_load(ymlfile)
    task_queue_config["tasks"]["diagnostic_reports"]["reports"] = diag_reports
    docker_util.string_into_container(yaml.dump(task_queue_config), container, "/home/worker/config/config.yml")
    docker_util.exec_safely(container, ["chown", "worker:worker", "/home/worker/config/config.yml"], user="root")


def fake_smtp_container(cfg):
    name = cfg.containers["fake_smtp"]
    return constellation.ConstellationContainer(name, cfg.fake_smtp_ref, ports=[1025, 1080])


def db_container(cfg):
    name = cfg.containers["db"]
    mounts = [constellation.ConstellationVolumeMount("db", "/pgdata")]
    return constellation.ConstellationContainer(name, cfg.db_ref, mounts=mounts, ports=[5432], configure=db_configure)


def db_configure(container, cfg):
    print("[db] Waiting for the database to accept connections")
    docker_util.exec_safely(container, ["montagu-wait.sh", "7200"])
    print("[db] Scrambling root password")
    db_set_root_password(container, cfg, cfg.db_root_password)
    print("[db] Setting up database users")
    db_setup_users(cfg)
    print("[db] Migrating database schema")
    db_migrate_schema(cfg)
    print("[db] Refreshing user permissions")
    # The migrations may have added new tables, so we should set the permissions
    # again, in case users need to have permissions on these new tables
    db_set_user_permissions(cfg)

    if cfg.enable_streaming_replication:
        print("[db] Enabling streaming replication")
        db_enable_streaming_replication(container, cfg)


def db_set_root_password(container, cfg, password):
    query = f"ALTER USER {cfg.db_root_user} WITH PASSWORD '{password}'"
    docker_util.exec_safely(container, f'psql -U {cfg.db_root_user} -d postgres -c "{query}"')


def db_setup_users(cfg):
    with db_connection(cfg) as conn:
        with conn.cursor() as cur:
            for user in cfg.db_users:
                database.setup_db_user(cur, user, cfg.db_users[user])
        conn.commit()


def db_set_user_permissions(cfg):
    with db_connection(cfg) as conn:
        with conn.cursor() as cur:
            for user in cfg.db_users:
                database.set_permissions(cur, user, cfg.db_users[user])
                # Revoke specific permissions now that all tables have been created.
                database.revoke_write_on_protected_tables(cur, user, cfg.db_protected_tables)
        conn.commit()


def db_connection(cfg, retries=5):
    attempt = 0
    while attempt < retries:
        try:
            return connect(
                user=cfg.db_root_user, dbname="montagu", password=cfg.db_root_password, host="localhost", port=5432
            )
        except OperationalError as _:
            attempt += 1
            if attempt == retries:
                raise
            print("Connection failed. Retrying in 5 seconds")
            time.sleep(5)


def db_migrate_schema(cfg):
    network_name = cfg.network
    image = cfg.db_migrate_ref
    client = docker.client.from_env()
    client.containers.run(
        str(image),
        [f"-user={cfg.db_root_user}", f"-password={cfg.db_root_password}", "migrate"],
        network=network_name,
        stderr=True,
        remove=True,
    )


def db_enable_streaming_replication(container, cfg):
    docker_util.exec_safely(
        container,
        ["enable-replication.sh", cfg.db_users["barman"]["password"], cfg.db_users["streaming_barman"]["password"]],
    )


def api_container(cfg):
    name = cfg.containers["api"]
    mounts = [
        constellation.ConstellationVolumeMount("burden_estimates", "/upload_dir"),
        constellation.ConstellationVolumeMount("emails", "/tmp/emails"),  # noqa S108
    ]
    return constellation.ConstellationContainer(name, cfg.api_ref, mounts=mounts, configure=api_configure)


def api_configure(container, cfg):
    print("[api] Configuring API container")
    docker_util.exec_safely(container, ["mkdir", "-p", "/etc/montagu/api"])
    inject_api_config(container, cfg)
    start_api(container)


def start_api(container):
    docker_util.exec_safely(container, ["touch", "/etc/montagu/api/go_signal"])


def inject_api_config(container, cfg):
    db_name = cfg.containers["db"]
    opts = {
        "app.url": f"https://{cfg.hostname}/api",
        "db.host": db_name,
        "db.username": "api",
        "db.password": cfg.db_users["api"]["password"],
        "allow.localhost": False,
        "celery.flower.host": cfg.containers["flower"],
        "packit.api.url": cfg.packit_api_url,
        "upload.dir": "/upload_dir",
    }

    if cfg.real_emails:
        opts["email.mode"] = "real"
        opts["email.password"] = cfg.email_password
        opts["flow.url"] = cfg.email_flow_url

    txt = "".join([f"{k}={v}\n" for k, v in opts.items()])
    docker_util.string_into_container(txt, container, "/etc/montagu/api/config.properties")


def proxy_container(cfg):
    name = cfg.containers["proxy"]
    proxy_ports = [cfg.proxy_port_http, cfg.proxy_port_https, cfg.proxy_port_metrics]

    mounts = []

    if cfg.use_acme:
        mounts += [constellation.ConstellationVolumeMount("montagu-tls", "/run/proxy")]

    return constellation.ConstellationContainer(
        name,
        cfg.proxy_ref,
        ports=proxy_ports,
        args=[str(cfg.proxy_port_https), cfg.hostname],
        mounts=mounts,
    )


def proxy_metrics_container(cfg):
    proxy_name = cfg.containers["proxy"]
    return constellation.ConstellationContainer(
        cfg.containers["metrics"],
        cfg.proxy_metrics_ref,
        ports=[9113],
        args=["-nginx.scrape-uri", f"http://{proxy_name}/basic_status"],
    )



import pytest

from src.montagu_deploy.config import MontaguConfig


def test_config_basic():
    cfg = MontaguConfig("config/basic")
    assert cfg.network == "montagu-network"
    assert cfg.volumes["db"] == "db_volume"
    assert cfg.volumes["emails"] == "emails"
    assert cfg.volumes["burden_estimates"] == "burden_estimate_files"
    assert cfg.volumes["guidance"] == "guidance_volume"
    assert cfg.volumes["templates"] == "template_volume"
    assert cfg.volumes["mq"] == "mq"
    assert cfg.container_prefix == "montagu"

    assert len(cfg.containers) == 10

    assert cfg.containers["api"] == "api"
    assert cfg.containers["db"] == "db"
    assert cfg.containers["admin"] == "admin"
    assert cfg.containers["contrib"] == "contrib"
    assert cfg.containers["proxy"] == "proxy"
    assert cfg.containers["metrics"] == "proxy-metrics"
    assert cfg.containers["mq"] == "mq"
    assert cfg.containers["flower"] == "flower"
    assert cfg.containers["task_queue"] == "task-queue"
    assert cfg.containers["fake_smtp"] == "fake-smtp"

    assert len(cfg.images) == 12

    assert str(cfg.images["db"]) == "ghcr.io/vimc/montagu-db:main"
    assert str(cfg.images["api"]) == "ghcr.io/vimc/montagu-api:main"
    assert str(cfg.images["admin"]) == "ghcr.io/vimc/montagu-admin-portal:main"
    assert str(cfg.images["api_admin"]) == "ghcr.io/vimc/montagu-cli:main"
    assert str(cfg.images["contrib"]) == "ghcr.io/vimc/montagu-contrib-portal:main"
    assert str(cfg.images["proxy"]) == "ghcr.io/vimc/montagu-proxy:main"
    assert str(cfg.images["mq"]) == "docker.io/redis:latest"
    assert str(cfg.images["flower"]) == "mher/flower:0.9.5"
    assert str(cfg.images["task_queue"]) == "ghcr.io/vimc/task-queue-worker:main"
    assert str(cfg.images["metrics"]) == "nginx/nginx-prometheus-exporter:1.3.0"
    assert str(cfg.images["db_migrate"]) == "ghcr.io/vimc/montagu-migrate:main"
    assert str(cfg.images["fake_smtp"]) == "reachfive/fake-smtp-server:latest"

    assert cfg.mq_port == 6379
    assert cfg.flower_port == 5555

    assert cfg.protect_data is False
    assert cfg.ssl_mode == "self-signed"

    assert cfg.db_root_user == "vimc"
    assert len(cfg.db_root_password) == 50
    assert cfg.db_users == {
        "api": {"password": "apipassword", "permissions": "all"},
        "import": {"password": "importpassword", "permissions": "all"},
        "orderly": {"password": "orderlypassword", "permissions": "all"},
        "readonly": {"password": "readonlypassword", "permissions": "readonly"},
    }
    assert len(cfg.db_protected_tables) == 12
    assert cfg.db_protected_tables[0] == "gavi_support_level"

    assert cfg.proxy_port_metrics == 9000


def test_config_email():
    cfg = MontaguConfig("config/basic")
    assert not cfg.real_emails

    options = {"api": {"email": {"password": "changeme"}}}
    with pytest.raises(Exception, match="flow_url"):
        MontaguConfig("config/basic", options=options)

    cfg = MontaguConfig("config/complete")
    assert cfg.real_emails
    assert cfg.email_password == "changeme"
    assert cfg.email_flow_url == "fakeurl"


def test_config_acme():
    cfg = MontaguConfig("config/acme")
    assert cfg.ssl_mode == "acme"
    acme_cfg = cfg.acme_config
    assert acme_cfg.email == "admin@montagu.org"


def test_config_generates_root_db_password():
    cfg = MontaguConfig("config/complete")
    assert cfg.db_root_password == "changeme"
    cfg = MontaguConfig("config/basic")
    assert cfg.db_root_password != "changeme"
    assert len(cfg.db_root_password) == 50


def test_config_streaming_replication():
    cfg = MontaguConfig("config/basic")
    assert not cfg.enable_streaming_replication
    cfg = MontaguConfig("config/complete")
    assert cfg.enable_streaming_replication
    assert cfg.db_users["barman"] == {"password": "barmanpassword", "option": "superuser"}
    assert cfg.db_users["streaming_barman"] == {"password": "streamingpassword", "option": "replication"}


def test_config_validates_db_user_permissions():
    options = {"db": {"users": {"api": {"permissions": "bad", "password": "pw"}}}}
    with pytest.raises(Exception, match="Invalid database permissions for 'api'"):
        MontaguConfig("config/basic", options=options)



"""Usage:
  montagu --version
  montagu configure <path>
  montagu start [--name=PATH] [--extra=PATH] [--option=OPTION]... [--pull]
  montagu status [--name=PATH]
  montagu stop [--name=PATH] [--volumes] [--network] [--kill] [--force]
    [--extra=PATH] [--option=OPTION]...

Options:
  --name=PATH      Override the configured name, use with care!
  --extra=PATH     Path, relative to <path>, of yml file of additional
                   configuration
  --option=OPTION  Additional configuration options, in the form key=value
                   Use dots in key for hierarchical structure, e.g., a.b=value
                   This argument may be repeated to provide multiple arguments
  --pull           Pull images before starting
  --volumes        Remove volumes (WARNING: irreversible data loss)
  --network        Remove network
  --kill           Kill the containers (faster, but possible db corruption)
"""

from pathlib import Path

import docopt
import yaml

import montagu_deploy.__about__ as about
from montagu_deploy.config import MontaguConfig
from montagu_deploy.montagu_constellation import montagu_constellation


def main(argv=None):
    path, extra, options, args = parse_args(argv)
    if args.version:
        print(about.__version__)
    else:
        if args.action == "configure":
            montagu_configure(path)
            return

        path = _read_identity(path)
        cfg = MontaguConfig(path, extra, options)
        obj = montagu_constellation(cfg)
        if args.action == "start":
            montagu_start(obj, args)
        elif args.action == "status":
            montagu_status(obj)
        elif args.action == "stop":
            montagu_stop(obj, args, cfg)


def parse_args(argv=None):
    opts = docopt.docopt(__doc__, argv)
    path = opts["<path>"] if opts["configure"] else opts["--name"]
    extra = opts["--extra"]
    options = parse_option(opts)
    return path, extra, options, MontaguArgs(opts)


def montagu_configure(name):
    prev = _read_identity(required=False)
    if prev:
        if prev != name:
            msg = (
                f"This montagu instance is already configured as '{prev}', "
                f"but you are trying to reconfigure it as '{name}'. "
                "If you really want to do do this, then delete the file "
                "'{IDENTITY_FILE}' from this directory and try again"
            )
            raise Exception(msg)
        else:
            print(f"Montagu already configured as '{name}")
    else:
        # Check that we can read the configuration before saving it.
        MontaguConfig(name)
        with IDENTITY_FILE.open("w") as f:
            f.write(name)
        print(f"Configured montagu as '{name}")


def montagu_start(obj, args):
    obj.start(pull_images=args.pull)


def montagu_status(obj):
    obj.status()


def montagu_stop(obj, args, cfg):
    if args.volumes:
        verify_data_loss(cfg)
    obj.stop(kill=args.kill, remove_network=args.network, remove_volumes=args.volumes)


def verify_data_loss(cfg):
    if cfg.protect_data:
        err = "Cannot remove volumes with this configuration"
        raise Exception(err)
    else:
        print(
            """WARNING! PROBABLE IRREVERSIBLE DATA LOSS!
You are about to delete the data volumes. This action cannot be undone
and will result in the irreversible loss of *all* data associated with
the application. This includes all databases etc."""
        )
    if not prompt_yes_no():
        msg = "Not continuing"
        raise Exception(msg)


def prompt_yes_no(get_input=input):
    return get_input("\nContinue? [yes/no] ") == "yes"


def parse_option(args):
    return [string_to_dict(x) for x in args["--option"]]


def string_to_dict(string):
    """Convert a configuration option a.b.c=x to a dictionary
    {"a": {"b": "c": x}}"""
    # Won't deal with dots embedded within quotes but that's ok as
    # that should not be allowed generally.
    try:
        key, value = string.split("=")
    except ValueError as err:
        msg = f"Invalid option '{string}', expected option in form key=value"
        raise Exception(msg) from err
    value = yaml_atom_parse(value)
    for k in reversed(key.split(".")):
        value = {k: value}
    return value


def yaml_atom_parse(x):
    ret = yaml.safe_load(x)
    if type(ret) not in [bool, int, float, str]:
        msg = f"Invalid value '{x}' - expected simple type"
        raise Exception(msg)
    return ret


class MontaguArgs:
    def __init__(self, args):
        if args["start"]:
            self.action = "start"
        elif args["status"]:
            self.action = "status"
        elif args["stop"]:
            self.action = "stop"
        elif args["configure"]:
            self.action = "configure"

        self.pull = args["--pull"]
        self.kill = args["--kill"]
        self.volumes = args["--volumes"]
        self.network = args["--network"]
        self.version = args["--version"]


IDENTITY_FILE = Path(".montagu_identity")


def _read_identity(name=None, *, required=True):
    if name:
        return name
    if IDENTITY_FILE.exists():
        with IDENTITY_FILE.open() as f:
            return f.read().strip()
    if required:
        msg = "Montagu identity is not yet configured; run 'montagu configure <name>' first"
        raise Exception(msg)
    return None



import os
from unittest import mock

import celery
import docker
import requests
import tenacity
import vault_dev
from constellation import docker_util
from packit_deploy.config import PackitConfig
from packit_deploy.packit_constellation import PackitConstellation
from YTClient.YTClient import YTClient
from YTClient.YTDataClasses import Command

from src.montagu_deploy import cli
from src.montagu_deploy.config import MontaguConfig
from tests import admin
from tests.utils import http_get


def test_start_stop_status():
    path = "config/basic"
    try:
        # Start
        cli.main(["start", "--name", path])

        cl = docker.client.from_env()
        containers = cl.containers.list()
        assert len(containers) == 10

        cfg = MontaguConfig(path)
        assert docker_util.network_exists(cfg.network)
        assert docker_util.volume_exists(cfg.volumes["db"])

        # Status
        cli.main(["status", "--name", "config/basic"])

        # Stop
        with mock.patch("src.montagu_deploy.cli.prompt_yes_no") as prompt:
            prompt.return_value = True
            cli.main(["stop", "--name", path, "--kill", "--volumes", "--network"])
            containers = cl.containers.list()
            assert len(containers) == 0
            assert not docker_util.network_exists(cfg.network)
            assert not docker_util.volume_exists(cfg.volumes["db"])
    finally:
        with mock.patch("src.montagu_deploy.cli.prompt_yes_no") as prompt:
            prompt.return_value = True
            cli.main(["stop", "--name", path, "--kill", "--volumes", "--network"])


def test_task_queue():
    packit_config_path = "tests"
    path = "config/ci"
    cfg = MontaguConfig(path)
    packit_config = PackitConfig(packit_config_path)
    try:
        youtrack_token = os.environ["YOUTRACK_TOKEN"]
        with vault_dev.Server(export_token=True) as s:
            cl = s.client()
            cl.write("secret/youtrack/token", value=youtrack_token)
            vault_addr = f"http://localhost:{s.port}"

            packit = PackitConstellation(packit_config)
            packit.start(pull_images=True)
            cli.main(
                [
                    "start",
                    "--name",
                    path,
                    f"--option=vault.addr={vault_addr}",
                    f"--option=vault.auth.args.token={s.token}",
                ]
            )

            # wait for APIs to be ready
            http_get("https://localhost/api/v1")
            wait_for_packit_api()

            add_task_queue_user(cfg, packit)
            app = celery.Celery(broker="redis://localhost//", backend="redis://")
            sig = "run-diagnostic-reports"
            args = ["testGroup", "testDisease", "testTouchstone-1", "2020-11-04T12:21:15", "no_vaccination"]
            signature = app.signature(sig, args)
            versions = signature.delay().get()

            assert len(versions) == 1
            # check expected notification email was sent to fake smtp server
            emails = requests.get("http://localhost:1080/api/emails", timeout=5).json()
            assert len(emails) == 1
            subj = "VIMC diagnostic report: testTouchstone-1 - testGroup - testDisease"
            assert emails[0]["subject"] == subj
            assert emails[0]["to"]["value"][0]["address"] == "minimal_modeller@example.com"
            if False:
                # skip everywhere because these tests are well meaning but don't work
                yt = YTClient("https://mrc-ide.myjetbrains.com/youtrack/", token=youtrack_token)
                issues = yt.get_issues("tag: {}".format("testTouchstone-1"))
                assert len(issues) == 1
                yt.run_command(Command(issues, "delete"))
    finally:
        with mock.patch("src.montagu_deploy.cli.prompt_yes_no") as prompt:
            prompt.return_value = True
            PackitConstellation(packit_config).stop(kill=True)
            cli.main(["stop", "--name", path, "--kill", "--volumes", "--network"])


def test_metrics_endpoints():
    packit_config_path = "tests"
    path = "config/basic"
    packit_config = PackitConfig(packit_config_path)
    try:
        packit = PackitConstellation(packit_config)
        packit.start(pull_images=True)
        cli.main(["start", "--name", path])

        wait_for_packit_api()

        assert "jvm_info" in http_get("http://localhost:9000/metrics/packit-api")
        assert "http_requests_duration_seconds_bucket" in http_get("http://localhost:9000/metrics/outpack_server")
    finally:
        with mock.patch("src.montagu_deploy.cli.prompt_yes_no") as prompt:
            prompt.return_value = True
            PackitConstellation(packit_config).stop(kill=True)
            cli.main(["stop", "--name", path, "--kill", "--volumes", "--network"])


@tenacity.retry(wait=tenacity.wait_fixed(1), stop=tenacity.stop_after_attempt(60))
def wait_for_packit_api():
    print("Trying to connect to packit api..")
    http_get("https://localhost/packit/api/auth/config")
    print("Success!")


def add_task_queue_user(cfg, packit):
    packit_db_container = packit.obj.containers.get("packit-db", "montagu")
    docker_util.exec_safely(
        packit_db_container,
        [
            "create-preauth-user",
            "--username",
            "task.queue",
            "--email",
            "montagu-task@imperial.ac.uk",
            "--displayname",
            "task.queue",
            "--role",
            "ADMIN",
        ],
    )

    admin.add_user(cfg, "task.queue", "task.queue", "montagu-task@imperial.ac.uk", "password")
    admin.add_role_to_user(cfg, "task.queue", "user")



# Methods for working with a database cursor object


def setup_db_user(curr, user, settings):
    create_user(curr, user, settings)
    set_password(curr, user, settings)
    set_permissions(curr, user, settings)


def create_user(db, user, settings):
    if "option" in settings:
        option = settings["option"]
    else:
        option = ""
    sql = """DO
    $body$
    BEGIN
       IF NOT EXISTS (SELECT FROM pg_catalog.pg_user WHERE usename = '{name}') THEN
          CREATE ROLE {name} {option} LOGIN PASSWORD '{password}';
       END IF;
    END
    $body$""".format(
        name=user, password=settings["password"], option=option
    )
    db.execute(sql)


def set_password(db, user, settings):
    db.execute(f"ALTER USER {user} WITH PASSWORD '{settings['password']}'")


def set_permissions(curr, user, settings):
    if "permissions" not in settings:
        return
    revoke_all(curr, user)
    permissions = settings["permissions"]
    if permissions == "all":
        grant_all(curr, user)
    elif permissions == "readonly":
        grant_readonly(curr, user)


def revoke_all(db, user):
    def revoke_all_on(what):
        db.execute(f"REVOKE ALL PRIVILEGES ON ALL {what} IN SCHEMA public FROM {user}")

    revoke_all_on("tables")
    revoke_all_on("sequences")
    revoke_all_on("functions")


def revoke_write_on_protected_tables(db, user, protected_tables):
    def revoke_specific_on(what):
        db.execute(f"REVOKE INSERT, UPDATE, DELETE ON {what} FROM {user}")

    for table in protected_tables:
        revoke_specific_on(table)


def grant_all(db, user):
    def grant_all_on(what):
        db.execute(f"GRANT ALL PRIVILEGES ON ALL {what} IN SCHEMA public TO {user}")

    print(f"  - Granting all permissions to {user}")
    grant_all_on("tables")
    grant_all_on("sequences")
    grant_all_on("functions")


def grant_readonly(db, user):
    print(f"  - Granting readonly permissions to {user}")
    db.execute(f"GRANT SELECT ON ALL TABLES IN SCHEMA public TO {user}")



import io
import os
import re
from contextlib import contextmanager, redirect_stdout
from unittest import mock

import pytest

from src.montagu_deploy import cli
from src.montagu_deploy.cli import prompt_yes_no, verify_data_loss
from src.montagu_deploy.config import MontaguConfig


@contextmanager
def transient_working_directory(path):
    origin = os.getcwd()
    try:
        if path is not None:
            os.chdir(path)
        yield
    finally:
        if path is not None:
            os.chdir(origin)


def test_parse_args():
    res = cli.parse_args(["start", "--name=config/basic", "--pull"])
    assert res[0] == "config/basic"
    assert res[1] is None
    assert res[2] == []
    args = res[3]
    assert args.action == "start"
    assert args.pull is True
    assert args.kill is False
    assert args.volumes is False
    assert args.network is False

    res = cli.parse_args(["start", "--name=config/basic", "--extra=extra.yml"])
    assert res[1] == "extra.yml"

    res = cli.parse_args(["start", "--name=config/basic", "--option=a=x", "--option=b.c=y"])
    assert res[2] == [{"a": "x"}, {"b": {"c": "y"}}]

    res = cli.parse_args(["stop", "--name=config/basic", "--kill", "--network", "--volumes"])
    args = res[3]
    assert args.action == "stop"
    assert args.pull is False
    assert args.kill is True
    assert args.volumes is True
    assert args.network is True

    res = cli.parse_args(["status", "--name=config/basic"])
    args = res[3]
    assert args.action == "status"

    res = cli.parse_args(["--version"])
    args = res[3]
    assert args.version is True


def test_version(capsys):
    cli.main(["--version"])
    out, _err = capsys.readouterr()
    assert re.match(r"\d+\.\d+\.\d+", out)


def test_args_passed_to_start():
    with mock.patch("src.montagu_deploy.cli.montagu_start") as f:
        cli.main(["start", "--name=config/basic"])

    assert f.called
    assert f.call_args[0][1].pull is False

    with mock.patch("src.montagu_deploy.cli.montagu_start") as f:
        cli.main(["start", "--name=config/basic", "--pull"])

    assert f.called
    assert f.call_args[0][1].pull is True


def test_args_passed_to_stop():
    with mock.patch("src.montagu_deploy.cli.montagu_stop") as f:
        cli.main(["stop", "--name=config/basic"])

    assert f.called
    assert f.call_args[0][1].kill is False
    assert f.call_args[0][1].network is False
    assert f.call_args[0][1].volumes is False

    with mock.patch("src.montagu_deploy.cli.montagu_stop") as f:
        cli.main(["stop", "--name=config/basic", "--volumes", "--network"])

    assert f.called
    assert f.call_args[0][1].kill is False
    assert f.call_args[0][1].network is True
    assert f.call_args[0][1].volumes is True


def test_args_passed_to_configure():
    with mock.patch("src.montagu_deploy.cli.montagu_configure") as f:
        cli.main(["configure", "config/basic"])

    assert f.call_count == 1
    assert f.mock_calls[0] == mock.call("config/basic")


def test_verify_data_loss_called():
    f = io.StringIO()
    with redirect_stdout(f):
        with mock.patch("src.montagu_deploy.cli.verify_data_loss") as verify:
            verify.return_value = True
            cli.main(["stop", "--name=config/basic", "--volumes"])

    assert verify.called


def test_verify_data_loss_not_called():
    f = io.StringIO()
    with redirect_stdout(f):
        with mock.patch("src.montagu_deploy.cli.verify_data_loss") as verify:
            verify.return_value = True
            cli.main(["stop", "--name=config/basic"])

    assert not verify.called


def test_verify_data_loss_warns_if_loss():
    cfg = MontaguConfig("config/basic")
    f = io.StringIO()
    with redirect_stdout(f):
        with mock.patch("src.montagu_deploy.cli.prompt_yes_no") as prompt:
            prompt.return_value = True
            verify_data_loss(cfg)

    assert prompt.called
    assert "WARNING! PROBABLE IRREVERSIBLE DATA LOSS!" in f.getvalue()


def test_verify_data_loss_throws_if_loss():
    cfg = MontaguConfig("config/basic")
    with mock.patch("src.montagu_deploy.cli.prompt_yes_no") as prompt:
        prompt.return_value = False
        with pytest.raises(Exception, match="Not continuing"):
            verify_data_loss(cfg)


def test_verify_data_prevents_unwanted_loss():
    cfg = MontaguConfig("config/basic")
    cfg.protect_data = True
    msg = "Cannot remove volumes with this configuration"
    with mock.patch("src.montagu_deploy.cli.prompt_yes_no"):
        with pytest.raises(Exception, match=msg):
            verify_data_loss(cfg)


def test_prompt_is_quite_strict():
    assert prompt_yes_no(lambda _: "yes")
    assert not prompt_yes_no(lambda _: "no")
    assert not prompt_yes_no(lambda _: "Yes")
    assert not prompt_yes_no(lambda _: "Great idea!")
    assert not prompt_yes_no(lambda _: "")


def test_bad_option_format():
    with pytest.raises(Exception, match="Invalid option"):
        cli.parse_args(["start", "--name=config/basic", "--option=one:two"])


def test_invalid_option_type():
    with pytest.raises(Exception, match="Invalid value"):
        cli.parse_args(["start", "--name=config/basic", "--option=one={2}"])


def test_can_read_identity(tmp_path):
    with transient_working_directory(tmp_path):
        assert cli._read_identity(required=False) is None
        assert cli._read_identity("foo") == "foo"
        with pytest.raises(Exception, match="not yet configured"):
            cli._read_identity()
        with open(".montagu_identity", "w") as f:
            f.write("foo\n")
        assert cli._read_identity() == "foo"


def test_can_configure_montagu(tmp_path, mocker):
    with transient_working_directory(tmp_path):
        mocker.patch("src.montagu_deploy.cli.MontaguConfig")
        cli.montagu_configure("foo")
        assert cli._read_identity() == "foo"
        cli.montagu_configure("foo")
        assert cli._read_identity() == "foo"
        with pytest.raises(Exception, match="already configured as 'foo'"):
            cli.montagu_configure("bar")
        assert cli._read_identity() == "foo"



import docker
from constellation import docker_util
from cryptography import x509
from cryptography.hazmat.backends import default_backend

from src.montagu_deploy.config import MontaguConfig
from src.montagu_deploy.montagu_constellation import montagu_constellation
from tests.utils import http_get


def test_start_and_stop():
    cfg = MontaguConfig("config/basic")
    obj = montagu_constellation(cfg)

    try:
        obj.start()

        cl = docker.client.from_env()

        assert docker_util.network_exists(cfg.network)
        assert docker_util.volume_exists(cfg.volumes["db"])
        assert docker_util.volume_exists(cfg.volumes["burden_estimates"])
        assert docker_util.volume_exists(cfg.volumes["emails"])
        assert docker_util.volume_exists(cfg.volumes["mq"])
        assert docker_util.volume_exists(cfg.volumes["templates"])
        assert docker_util.volume_exists(cfg.volumes["guidance"])

        assert docker_util.container_exists("montagu-api")
        assert docker_util.container_exists("montagu-db")
        assert docker_util.container_exists("montagu-proxy")
        assert docker_util.container_exists("montagu-proxy-metrics")
        assert docker_util.container_exists("montagu-admin")
        assert docker_util.container_exists("montagu-contrib")
        assert docker_util.container_exists("montagu-mq")
        assert docker_util.container_exists("montagu-flower")
        assert docker_util.container_exists("montagu-task-queue")
        assert docker_util.container_exists("montagu-fake-smtp")

        containers = cl.containers.list()
        print([c.name for c in containers])
        assert len(containers) == 10

    finally:
        obj.stop(kill=True, remove_volumes=True)


def test_api_configured():
    cfg = MontaguConfig("config/basic")
    obj = montagu_constellation(cfg)

    try:
        obj.start()

        api = get_container(cfg, "api")
        api_config = docker_util.string_from_container(api, "/etc/montagu/api/config.properties").split("\n")

        assert "app.url=https://localhost/api" in api_config
        assert "db.host=db" in api_config
        assert "db.username=api" in api_config
        assert "db.password=apipassword" in api_config
        assert "allow.localhost=False" in api_config
        assert "upload.dir=/upload_dir" in api_config
        assert "email.mode=real" not in api_config

        res = http_get("https://localhost/api/v1")
        assert '"status": "success"' in res
    finally:
        obj.stop(kill=True, remove_volumes=True)

    cfg = MontaguConfig("config/complete")
    obj = montagu_constellation(cfg)

    try:
        obj.start()
        api = get_container(cfg, "api")
        api_config = docker_util.string_from_container(api, "/etc/montagu/api/config.properties").split("\n")
        assert "email.mode=real" in api_config
        assert "email.password=changeme" in api_config
        assert "flow.url=fakeurl" in api_config

    finally:
        obj.stop(kill=True, remove_volumes=True)


def test_proxy_configured_self_signed():
    cfg = MontaguConfig("config/basic")
    obj = montagu_constellation(cfg)

    try:
        obj.start()

        api = get_container(cfg, "proxy")
        cert = docker_util.string_from_container(api, "/etc/montagu/proxy/certificate.pem")
        key = docker_util.string_from_container(api, "/etc/montagu/proxy/ssl_key.pem")
        assert cert is not None
        assert key is not None

        res = http_get("https://localhost")
        assert "Montagu" in res

    finally:
        obj.stop(kill=True, remove_volumes=True)


def test_db_configured():
    cfg = MontaguConfig("config/complete")
    obj = montagu_constellation(cfg)

    try:
        obj.start()

        db = get_container(cfg, "db")
        res = docker_util.exec_safely(db, f'psql -U {cfg.db_root_user} -d postgres -c "\\du"')
        res = res.output.decode("UTF-8")

        for u in cfg.db_users:
            assert u in res

        query = "SELECT * FROM pg_replication_slots WHERE slot_name = 'barman'"
        res = docker_util.exec_safely(db, f'psql -U {cfg.db_root_user} -d postgres -c "{query}"')
        res = res.output.decode("UTF-8")

        assert "barman" in res

    finally:
        obj.stop(kill=True, remove_volumes=True)


def test_proxy_configured_ssl():
    cfg = MontaguConfig("config/complete")
    obj = montagu_constellation(cfg)

    try:
        obj.start()
        proxy = get_container(cfg, "proxy")
        cert_str = docker_util.string_from_container(proxy, "/etc/montagu/proxy/certificate.pem")
        cert = x509.load_pem_x509_certificate(cert_str.encode(), default_backend())
        assert cert.subject == cert.issuer

    finally:
        obj.stop(kill=True, remove_volumes=True)


def test_metrics():
    cfg = MontaguConfig("config/basic")
    obj = montagu_constellation(cfg)

    try:
        obj.start()
        http_get("http://localhost:9113/metrics")

    finally:
        obj.stop(kill=True)


def get_container(cfg, name):
    cl = docker.client.from_env()
    return cl.containers.get(f"{cfg.container_prefix}-{cfg.containers[name]}")



import docker


def add_user(cfg, name, username, email, password):
    args = ["add", name, username, email, password, "--if-not-exists"]
    return run(cfg, args)


def add_role_to_user(cfg, name, role):
    args = ["addRole", name, role]
    return run(cfg, args)


def run(cfg, args):
    image = str(cfg.images["api_admin"])
    client = docker.client.from_env()
    result = client.containers.run(image, args, network=cfg.network, stderr=True)
    print(result.decode("UTF-8"))



import random
import string

import constellation
from constellation import config


class MontaguConfig:
    def __init__(self, path, extra=None, options=None):
        dat = config.read_yaml(f"{path}/montagu.yml")
        dat = config.config_build(path, dat, extra, options)
        self.path = path
        self.vault = config.config_vault(dat, ["vault"])
        self.network = config.config_string(dat, ["network"])
        self.protect_data = config.config_boolean(dat, ["protect_data"])
        self.volumes = config.config_dict(dat, ["volumes"])

        self.container_prefix = config.config_string(dat, ["container_prefix"])
        self.repo = config.config_string(dat, ["repo"])
        self.hostname = config.config_string(dat, ["hostname"])
        self.packit_api_url = config.config_string(dat, ["packit_api_url"])

        # API
        self.api_ref = self.build_ref(dat, "api")
        self.api_admin_ref = self.build_ref(dat["api"], "admin")
        self.real_emails = "email" in dat["api"]
        if self.real_emails:
            self.email_password = config.config_string(dat, ["api", "email", "password"])
            self.email_flow_url = config.config_string(dat, ["api", "email", "flow_url"])

        # DB
        self.db_ref = self.build_ref(dat, "db")
        self.db_migrate_ref = self.build_ref(dat["db"], "migrate")
        self.db_root_user = config.config_string(dat, ["db", "root_user"])
        if "root_password" in dat["db"]:
            self.db_root_password = config.config_string(dat, ["db", "root_password"])
        else:
            self.db_root_password = "".join(
                random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(50)
            )
        self.db_users = config.config_dict(dat, ["db", "users"])
        invalid = {
            k: v for k, v in self.db_users.items() if "permissions" in v and v["permissions"] not in ["all", "readonly"]
        }
        if any(invalid):
            invalid_str = ",".join(iter(invalid.keys()))
            msg = f"Invalid database permissions for '{invalid_str}'. Supported values are 'all' and 'readonly'"
            raise Exception(msg)
        self.db_protected_tables = config.config_list(dat, ["db", "protected_tables"])
        self.enable_streaming_replication = "barman" in self.db_users and "streaming_barman" in self.db_users

        # Proxy
        self.proxy_ref = self.build_ref(dat, "proxy")
        self.proxy_port_http = config.config_integer(dat, ["proxy", "port_http"])
        self.proxy_port_https = config.config_integer(dat, ["proxy", "port_https"])
        self.proxy_port_metrics = config.config_integer(dat, ["proxy", "port_metrics"], is_optional=True, default=9000)
        self.proxy_metrics_ref = self.build_ref(dat["proxy"], "metrics")

        # Portals
        self.admin_ref = self.build_ref(dat, "admin")
        self.contrib_ref = self.build_ref(dat, "contrib")

        # Task Q
        self.mq_ref = self.build_ref(dat, "mq")
        self.mq_port = config.config_integer(dat, ["mq", "port"])
        self.flower_ref = self.build_ref(dat, "flower")
        self.flower_port = config.config_integer(dat, ["flower", "port"])
        self.task_queue_ref = self.build_ref(dat, "task_queue")
        self.task_queue_tasks = config.config_dict(dat, ["task_queue", "tasks"])
        self.task_queue_servers = config.config_dict(dat, ["task_queue", "servers"])
        if "fake_smtp_server" in dat:
            self.fake_smtp_ref = self.build_ref(dat, "fake_smtp_server")
        else:
            self.fake_smtp_ref = False

        self.containers = {
            "db": "db",
            "api": "api",
            "proxy": "proxy",
            "metrics": "proxy-metrics",
            "admin": "admin",
            "contrib": "contrib",
            "mq": "mq",
            "flower": "flower",
            "task_queue": "task-queue",
        }

        if self.fake_smtp_ref:
            self.containers["fake_smtp"] = "fake-smtp"

        self.images = {
            "db": self.db_ref,
            "api": self.api_ref,
            "proxy": self.proxy_ref,
            "metrics": self.proxy_metrics_ref,
            "admin": self.admin_ref,
            "contrib": self.contrib_ref,
            "mq": self.mq_ref,
            "flower": self.flower_ref,
            "task_queue": self.task_queue_ref,
            "api_admin": self.api_admin_ref,
            "db_migrate": self.db_migrate_ref,
        }

        if self.fake_smtp_ref:
            self.images["fake_smtp"] = self.fake_smtp_ref

        # Acme-Buddy
        acme_key = "acme_buddy"
        self.use_acme = acme_key in dat
        if self.use_acme:
            self.acme_config = config.config_acme(dat, acme_key)
            self.containers["acme-buddy"] = "acme-buddy"
            self.images["acme-buddy"] = self.acme_config.ref
            self.volumes["montagu-tls"] = "montagu-tls"
            self.ssl_mode = "acme"
        else:
            self.ssl_mode = "self-signed"

    def build_ref(self, dat, section):
        name = config.config_string(dat, [section, "name"])
        tag = config.config_string(dat, [section, "tag"])
        if "repo" in dat[section]:
            repo = config.config_string(dat, [section, "repo"])
        else:
            repo = self.repo
        return constellation.ImageReference(repo, name, tag)
